// Copyright 2017-2023 Azul Systems, Inc.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//
// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright notice,
// this list of conditions and the following disclaimer in the documentation
// and/or other materials provided with the distribution.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

package jdk.internal.crac;

import jdk.internal.access.JavaIOFileDescriptorAccess;
import jdk.internal.access.SharedSecrets;

import java.io.FileDescriptor;
import java.util.List;
import java.util.Objects;
import java.util.WeakHashMap;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;


public class ClaimedFDs {
    private final static JavaIOFileDescriptorAccess fileDescriptorAccess = SharedSecrets.getJavaIOFileDescriptorAccess();

    private WeakHashMap<FileDescriptor, Descriptor> fds = new WeakHashMap<>();

    public static class Descriptor {
        private int fd;
        private final Object claimer;
        private final Supplier<Exception> exceptionSupplier;

        public Descriptor(Object claimer, Supplier<Exception> exceptionSupplier) {
            this.fd = -1;
            this.claimer = claimer;
            this.exceptionSupplier = exceptionSupplier;
        }

        void setFd(int fd) {
            assert this.fd == -1;
            this.fd = fd;
        }

        public int getFd() {
            assert this.fd != -1;
            return fd;
        }

        public Object getClaimer() {
            return claimer;
        }

        public Supplier<Exception> getExceptionSupplier() {
            return exceptionSupplier;
        }

        @Override
        public String toString() {
            return "{fd=" + fd + ", claimer=" + claimer + '}';
        }
    }

    public List<Descriptor> getClaimedFds() {
        List<Descriptor> list = fds.entrySet().stream()
            .filter((var e) -> e.getKey().valid())
            .map(entry -> {
                    Descriptor d = entry.getValue();
                    d.setFd(fileDescriptorAccess.get(entry.getKey()));
                    return d;
                })
            .collect(Collectors.toList());
        // destroy fds since we've modified Descriptors
        fds = null;
        return list;
    }

    public void claimFd(FileDescriptor fd, Object claimer, Supplier<Exception> supplier, Object... suppressedClaimers) {
        if (fd == null) {
            return;
        }

        Descriptor descriptor = fds.get(fd);
        LoggerContainer.debug("ClaimFD: fd {0} claimer {1} existing {2}",
            fd, claimer, descriptor != null ? descriptor.claimer : "NONE");
        if (descriptor == null ||
                Stream.of(suppressedClaimers).anyMatch((supressed) -> supressed == descriptor.getClaimer())) {
            fds.put(fd, new Descriptor(claimer, supplier));
        }
    }
}
